/**
 * @file		hmp_parser.c
 * @date		22/08/2022
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Process HMP file structure and extract its datas.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "RSPTerrain_errordefs.h"
#include "RSPTerrain_datatypes.h"
#include "hmp_struct.h"
#include "hmp_parser.h"


////////////////////////////////////////////////////////////////////////////////
//                       Private functions declarations
////////////////////////////////////////////////////////////////////////////////

static unsigned short extractTerrainHMP(T_RSPTERRAIN_HMP*, const MEMFILE, const RSPTERRAIN_PARAMETERS*);
static unsigned short resolveTilesMap(T_RSPTERRAIN_HMP* pHmpStruct, const T_TILE_INDICES* tiles_indices,
		const T_HMPFILE_TILE* tiles, const RSPTERRAIN_PARAMETERS* pParams);


////////////////////////////////////////////////////////////////////////////////
//                         Public functions definition
////////////////////////////////////////////////////////////////////////////////

unsigned short RSP_TerrainLib_ParseHMPMemFile(const MEMFILE pMemFile, T_RSPTERRAIN_HMP* hmpStruct, const RSPTERRAIN_PARAMETERS* pParams) {
	unsigned char err = RSPLIB_SUCCESS;

	if (hmpStruct != NULL && pMemFile != NULL) {
		// Do the magic!
		err = extractTerrainHMP(hmpStruct, pMemFile, pParams);
	} else err = RSPLIB_ERROR_ARGS_NULL;

	return err;
}

unsigned short RSP_TerrainLib_ParseHMPFile(const char* fileName, T_RSPTERRAIN_HMP* hmpStruct, const RSPTERRAIN_PARAMETERS* pParams) {
	unsigned char err = RSPLIB_SUCCESS;
	long fileSize;
	FILE* fStream = NULL;
	MEMFILE memFile = NULL;

	if (hmpStruct != NULL && fileName != NULL) {
		// Open file
		fStream = fopen(fileName, "rb");

		if (fStream != NULL) {
			// Determine file size in bytes
			fseek(fStream, 0, SEEK_END);
			fileSize = ftell(fStream);
			fseek(fStream, 0, SEEK_SET);
			if (pParams->verbose_mode) printf("[DBG] > Input file size: %ld bytes\n", fileSize);

			memFile = malloc(fileSize + 1);
			if (memFile != NULL) {
				// Copy file in RAM
				fread(memFile, fileSize, 1, fStream);
				fclose(fStream);

				// Do the magic!
				err = extractTerrainHMP(hmpStruct, memFile, pParams);

				free(memFile);

			} else {
				fclose(fStream);
				err = RSPLIB_ERROR_MEMORY;
				if (pParams->verbose_mode) printf("[ERR] Can't allocate enough memory for file processing!\n");
			}
		} else {
			err = RSPLIB_ERROR_IO;
			if (pParams->verbose_mode) printf("[ERR] Input file %s not found!\n", fileName);
		}
	} else err = RSPLIB_ERROR_ARGS_NULL;

	return err;
}


////////////////////////////////////////////////////////////////////////////////
//                       Private functions definition
////////////////////////////////////////////////////////////////////////////////

/**
 * @brief Extract datas from terrain HMP in memory.
 *
 * @param[out] pHmpStruct		Take T_RSPTERRAIN_HMP structure to fill with parsed datas.
 * @param[in] pMemfile			Pointer to an in-memory file location.
 * @param[in] pParams			Program option, used to tune parser features.
 *
 * @return Error code, RSPLIB_SUCCESS when no error.
 */
static unsigned short extractTerrainHMP(T_RSPTERRAIN_HMP* pHmpStruct, const MEMFILE pMemfile, const RSPTERRAIN_PARAMETERS* pParams) {
	unsigned short errorCode = RSPLIB_SUCCESS;
	T_TILE_INDICES* tiles_indices = NULL;
	T_HMPFILE_TILE* tiles = NULL;
	unsigned int tiles_offset = 0;

	if (pHmpStruct == NULL || pMemfile == NULL) return RSPLIB_ERROR_ARGS_NULL;

	// Get header infos
	pHmpStruct->height_scale = ((T_HMPFILE_HEADER *)pMemfile)->height_scale;
	tiles_offset = ((T_HMPFILE_HEADER *)pMemfile)->tiles_start_offset;
	pHmpStruct->tiles_count = ((T_HMPFILE_HEADER *)pMemfile)->tiles_count;
	pHmpStruct->width = ((T_HMPFILE_HEADER *)pMemfile)->width_BLK;
	pHmpStruct->height = ((T_HMPFILE_HEADER *)pMemfile)->height_BLK;

	if (pParams->verbose_mode) {
		printf("[DBG] > Height scale: %.8f\n", pHmpStruct->height_scale);
		printf("[DBG] > Tiles count: %d\n", pHmpStruct->tiles_count);
		printf("[DBG] > Tiles offset: 0x%X\n", tiles_offset);
		printf("[DBG] > Terrain size: %dx%d\n", pHmpStruct->width, pHmpStruct->height);
		printf("\n");
	}
	if (pParams->god_mode) {
		printf("[DBG] > Unknown0: %d\n", ((T_HMPFILE_HEADER *)pHmpStruct)->unknown0);
		printf("[DBG] > Unknown1: %d\n", ((T_HMPFILE_HEADER *)pHmpStruct)->unknown1);
		printf("\n");
	}

	// Get tiles indices
	tiles_indices = (T_TILE_INDICES *)malloc(pHmpStruct->width * pHmpStruct->height * sizeof(T_TILE_INDICES));
	memcpy(tiles_indices,
		pMemfile + sizeof(T_HMPFILE_HEADER),
		pHmpStruct->width * pHmpStruct->height * sizeof(unsigned short));

	// Get tiles datas
	tiles = (T_HMPFILE_TILE *)malloc(pHmpStruct->tiles_count * sizeof(T_HMPFILE_TILE));
	memcpy(tiles,
		pMemfile + tiles_offset,
		((T_HMPFILE_HEADER *)pMemfile)->tiles_count * sizeof(T_HMPFILE_TILE));

	if (tiles && tiles_indices) {
		// Reconstruct tiles map from tiles datas and tiles indices
		errorCode = resolveTilesMap(pHmpStruct, tiles_indices, tiles, pParams);
	} else errorCode = RSPLIB_ERROR_MEMORY;

	if (tiles) free(tiles);
	if (tiles_indices) free(tiles_indices);

	return errorCode;
}

/**
 * @brief Recreate original tilesmap using index and tiles in memory.
 *
 * @param[in|out]	pHmpStruct		Take T_RSPTERRAIN_HMP structure to fill with parsed datas.
 * @param[in]		tiles_indices	Tiles indices list.
 * @param[in]		tiles			Tiles datas list.
 * @param[in]		pParams			Program option, used to tune parser features.
 *
 * @return Error code, RSPLIB_SUCCESS when no error.
 */
static unsigned short resolveTilesMap(T_RSPTERRAIN_HMP* pHmpStruct, const T_TILE_INDICES* tiles_indices,
		const T_HMPFILE_TILE* tiles, const RSPTERRAIN_PARAMETERS* pParams) {

	T_TILE_INDICES tiles_idx;
	unsigned int i,j,k,l,remap_index;

	// Build tilesmap array to contain height values
	pHmpStruct->tilesmap = malloc(pHmpStruct->width * pHmpStruct->height * sizeof(T_RSPTERRAIN_TILE));
	if (pHmpStruct->tilesmap == NULL) return RSPLIB_ERROR_MEMORY;

	// Fetch tile from ID
	for ( i = 0; i < pHmpStruct->height; i++ ) {
		for ( j = 0; j < pHmpStruct->width; j++ ) {

			remap_index = pParams->invertZ ? i * pHmpStruct->height + (pHmpStruct->width-1 - j) : i * pHmpStruct->height + j;
			tiles_idx = tiles_indices[i * pHmpStruct->height + j];

			// Get tiles pseudodatas
			pHmpStruct->tilesmap[remap_index].texmap_id = tiles[tiles_idx].texmap_id;
			pHmpStruct->tilesmap[remap_index].low_height = tiles[tiles_idx].low_height;
			pHmpStruct->tilesmap[remap_index].high_height = tiles[tiles_idx].high_height;

			// Get the 5x5 bytes height values for this tile
			for ( k = 0; k < 5; k++ ) {
				for ( l = 0; l < 5; l++ ) {
					if (pParams->invertZ)
						pHmpStruct->tilesmap[remap_index].height_values[4-l][k] = pParams->negative ? tiles[tiles_idx].height_values[k][l] + 128 : 255 - (tiles[tiles_idx].height_values[k][l] + 128);
					else
						pHmpStruct->tilesmap[remap_index].height_values[l][k] = pParams->negative ? tiles[tiles_idx].height_values[k][l] + 128 : 255 - (tiles[tiles_idx].height_values[k][l] + 128);
				}
			}

		}
	}

	return RSPLIB_SUCCESS;
}
